/*===========================================================================
/
/									ttPage
/								GR[hϊ
/
/============================================================================
/ Copyright (C) 1998-2007 Sota. All rights reserved.
/
/ Redistribution and use in source and binary forms, with or without 
/ modification, are permitted provided that the following conditions 
/ are met:
/
/  1. Redistributions of source code must retain the above copyright 
/     notice, this list of conditions and the following disclaimer.
/  2. Redistributions in binary form must reproduce the above copyright 
/     notice, this list of conditions and the following disclaimer in the 
/     documentation and/or other materials provided with the distribution.
/
/ THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
/ IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
/ OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
/ IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, 
/ INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
/ BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF 
/ USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
/ ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
/ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
/ THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/============================================================================*/

#define  STRICT
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mbstring.h>
#include <windowsx.h>
#include <tchar.h>

#include "common.h"
#include "resource.h"


#define CONV_ASCII		0		/* ASCII */
#define CONV_KANJI		1		/*  */
#define CONV_KANA		2		/* pJ^Ji */

#define OUT_BUFFER_ADD		(4 * 1024)

/*===== vg^Cv =====*/

static int CheckOnSJIS(BYTE *pos, BYTE *btm);
static int CheckOnEUC(BYTE *pos, BYTE *btm);
static int CheckOnUTF8(BYTE *pos, BYTE *btm);


/*-------------------------------------------------------------------------------
Name    :	ConvEUCtoSJIS
Desc    :	EUCR[hSHIFT-JISR[hɕϊ
Param   :	cInfo	R[hϊ
Return  :	none
Note    :	
-------------------------------------------------------------------------------*/
void ConvEUCtoSJIS(CODECONVINFO *cInfo)
{
	int		code;
	BYTE	*get;
	BYTE	*put;
	int		rest;
	int		len;
	int		bufSize;
	BYTE	kanjiFirst;

	bufSize = cInfo->OrgLen + OUT_BUFFER_ADD;
	cInfo->Str = Malloc(bufSize);
	put = cInfo->Str;
	get = cInfo->OrgStr;
	len = 0;
	kanjiFirst = 0;

	for(rest = cInfo->OrgLen; rest > 0; rest--)
	{
		if((len + 2) > bufSize)
		{
			/* Resize output buffer */
			bufSize += OUT_BUFFER_ADD;
			cInfo->Str = Realloc(cInfo->Str, bufSize);
			put = cInfo->Str + len;
		}

		if((*get & 0x80) != 0)
		{
			if(kanjiFirst == 0)
			{
				kanjiFirst = *get++;
			}
			else
			{
				if(kanjiFirst == 0x8E)		/* pJ^Ji */
				{
					*put++ = *get++;
					len++;
				}
				else
				{
					code = _mbcjistojms(((kanjiFirst & 0x7F) * 0x100) + (*get++ & 0x7F));
					*put++ = HIGH8(code);
					*put++ = LOW8(code);
					len += 2;
				}
				kanjiFirst = 0;
			}
		}
		else
		{
			if(kanjiFirst != 0)
			{
				*put++ = kanjiFirst;
				len++;
				kanjiFirst = 0;
			}
			*put++ = *get++;
			len++;
		}
	}

	if(kanjiFirst != 0)
	{
		if((len + 1) > bufSize)
		{
			/* Resize output buffer */
			bufSize += 1;
			cInfo->Str = Realloc(cInfo->Str, bufSize);
			put = cInfo->Str + len;
		}
		*put = kanjiFirst;
		len++;
	}

	cInfo->Len = len;

	return;
}


/*-------------------------------------------------------------------------------
Name    :	ConvJIStoSJIS
Desc    :	JISR[hSHIFT-JISR[hɕϊ
Param   :	cInfo	R[hϊ
Return  :	none
Note    :	GXP[vR[h́Â̂ɑΉĂ
			Jn		<ESC>$B		<ESC>$@
			pJiJn	<ESC>(I
			I		<ESC>(B		<ESC>(J		<ESC>(H
-------------------------------------------------------------------------------*/
void ConvJIStoSJIS(CODECONVINFO *cInfo)
{
	int		code;
	BYTE	*get;
	BYTE	*put;
	int		rest;
	int		len;
	int		bufSize;
	BYTE	kanjiFirst;
	int		escProc;
	BYTE	escCode[2];
	int		kanjiMode;

	bufSize = cInfo->OrgLen + OUT_BUFFER_ADD;
	cInfo->Str = Malloc(bufSize);
	put = cInfo->Str;
	get = cInfo->OrgStr;
	len = 0;
	kanjiFirst = 0;
	escProc = 0;

	for(rest = cInfo->OrgLen; rest > 0; rest--)
	{
		if((len + 2) > bufSize)
		{
			/* Resize output buffer */
			bufSize += OUT_BUFFER_ADD;
			cInfo->Str = Realloc(cInfo->Str, bufSize);
			put = cInfo->Str + len;
		}

		if(escProc == 0)
		{
			if(*get == 0x1B)
			{
				if(kanjiFirst != 0)
				{
					*put++ = kanjiFirst;
					len++;
					kanjiFirst = 0;
				}
				escCode[escProc] = *get++;
				escProc++;
			}
			else
			{
				if(kanjiMode == CONV_KANA)
				{
					if(kanjiFirst != 0)
					{
						*put++ = kanjiFirst;
						len++;
						kanjiFirst = 0;
					}

					if((*get >= 0x21) && (*get <= 0x5F))
					{
						*put++ = *get++ + 0x80;
						len++;
					}
					else
					{
						*put++ = *get++;
						len++;
					}
				}
				else if(kanjiMode == CONV_KANJI)
				{
					if((*get >= 0x21) && (*get <= 0x7E))
					{
						if(kanjiFirst == 0)
						{
							kanjiFirst = *get++;
						}
						else
						{
							code = _mbcjistojms((kanjiFirst * 0x100) + *get++);
							*put++ = HIGH8(code);
							*put++ = LOW8(code);
							len += 2;
							kanjiFirst = 0;
						}
					}
					else
					{
						if(kanjiFirst == 0)
						{
							*put++ = *get++;
							len++;
						}
						else
						{
							*put++ = kanjiFirst;
							*put++ = *get++;
							len += 2;
							kanjiFirst = 0;
						}
					}
				}
				else
				{
					*put++ = *get++;
					len++;
				}
			}
		}
		else if(escProc == 1)
		{
			if((*get == _T('$')) || (*get == _T('(')))
			{
				escCode[escProc] = *get++;
				escProc++;
			}
			else
			{
				*put++ = escCode[0];
				*put++ = *get++;
				len += 2;
				escProc = 0;
			}
		}
		else if(escProc == 2)
		{
			if((escCode[1] == _T('$')) && ((*get == _T('B')) || (*get == _T('@'))))
			{
				kanjiMode = CONV_KANJI;
			}
			else if((escCode[1] == _T('(')) && (*get == _T('I')))
			{
				kanjiMode = CONV_KANA;
			}
			else if((escCode[1] == _T('(')) && ((*get == _T('B')) || (*get == _T('J')) || (*get == _T('H'))))
			{
				kanjiMode = CONV_ASCII;
			}
			else
			{
				*put++ = escCode[0];
				*put++ = escCode[1];
				len += 2;
				if((kanjiMode == CONV_KANJI) && ((*get >= 0x21) && (*get <= 0x7E)))
				{
					kanjiFirst = *get;
				}
				else
				{
					*put++ = *get;
					len++;
				}
			}
			get++;
			escProc = 0;
		}
	}

	if((len + 1 + escProc) > bufSize)
	{
		/* Resize output buffer */
		bufSize += 3;
		cInfo->Str = Realloc(cInfo->Str, bufSize);
		put = cInfo->Str + len;
	}

	if(kanjiFirst != 0)
	{
		*put++ = kanjiFirst;
		len++;
	}
	if(escProc >= 1)
	{
		*put++ = escCode[0];
		len++;
	}
	if(escProc == 2)
	{
		*put++ = escCode[1];
		len++;
	}

	cInfo->Len = len;

	return;
}


/*-------------------------------------------------------------------------------
Name    :	CheckEncoding
Desc    :	GR[h̎ނ𒲂ׂ
Param   :	text	
			size 	oCg
			encode	킩ĂGR[h
Return  :	GR[h̎ (ENCODE_xxx)
Note    :	
-------------------------------------------------------------------------------*/
int CheckEncoding(BYTE *text, int size, int encode)
{
	int		ret;
	BYTE	*btm;
	BYTE	*pos;
	BYTE	*tmp;
	int		pointSJIS;
	int		pointEUC;
	int		pointUTF8;

	switch(encode)
	{
		case ENCODE_UTF16_CHECK :
			if((size >= 4) && (*text == 0xFF) && (*(text+1) == 0xFE))
			{
				ret = ENCODE_UTF16_BOM;
			}
			else
			{
				ret = ENCODE_UTF16_NOBOM;
			}
			break;

		case ENCODE_UTF16_BE_CHECK :
			if((size >= 4) && (*text == 0xFE) && (*(text+1) == 0xFF))
			{
				ret = ENCODE_UTF16_BOM_BE;
			}
			else
			{
				ret = ENCODE_UTF16_NOBOM_BE;
			}
			break;

		case ENCODE_UTF8_CHECK :
			if((size >= 4) && (*text == 0xEF) && (*(text+1) == 0xBB) && (*(text+2) == 0xBF))
			{
				ret = ENCODE_UTF8_BOM;
			}
			else
			{
				ret = ENCODE_UTF8_NOBOM;
			}
			break;

		case ENCODE_AUTO :
			ret = ENCODE_UNKNOWN;
			if((size >= 4) && (*text == 0xFF) && (*(text+1) == 0xFE))
			{
				ret = ENCODE_UTF16_BOM;
			}
			else if((size >= 4) && (*text == 0xFE) && (*(text+1) == 0xFF))
			{
				ret = ENCODE_UTF16_BOM_BE;
			}
			else if((size >= 4) && (*text == 0xEF) && (*(text+1) == 0xBB) && (*(text+2) == 0xBF))
			{
				ret = ENCODE_UTF8_BOM;
			}
			else
			{
				btm = text + size;

				if (size >= 3)
				{
					/* JISR[h̃`FbN */
					pos = text;
					while ((pos = memchr(pos, 0x1b, btm - pos - 2)) != NULL)
					{
						pos++;
						if ((memcmp(pos, "$B", 2) == 0) ||	/* <ESC>$B */
							(memcmp(pos, "$@", 2) == 0) ||	/* <ESC>$@ */
							(memcmp(pos, "(I", 2) == 0))		/* <ESC>(I */
						{
							ret = ENCODE_JIS;
							break;
						}
					}
				}

				if((ret == ENCODE_UNKNOWN) && (size >= 2))
				{
					/* EUCSHIFT-JISR[h̃`FbN */
					ret = ENCODE_SJIS;
					pos = text;
					while(pos < btm)
					{
						pointSJIS = CheckOnSJIS(pos, btm);
						pointEUC = CheckOnEUC(pos, btm);
						pointUTF8 = CheckOnUTF8(pos, btm);
						if(pointSJIS > pointEUC)
						{
							if(pointSJIS > pointUTF8)
							{
								ret = ENCODE_SJIS;
								break;
							}
							if(pointSJIS < pointUTF8)
							{
								ret = ENCODE_UTF8_NOBOM;
								break;
							}
						}
						if(pointEUC > pointSJIS)
						{
							if(pointEUC > pointUTF8)
							{
								ret = ENCODE_EUC;
								break;
							}
							if(pointEUC < pointUTF8)
							{
								ret = ENCODE_UTF8_NOBOM;
								break;
							}
						}
						tmp = memchr(pos, '\n', btm-pos);
						if(tmp == NULL)
						{
							tmp = memchr(pos, '\r', btm-pos);
						}
						if(tmp == NULL)
						{
							break;
						}
						pos = tmp + 1;
					}
				}

				if(ret == ENCODE_UNKNOWN)
				{
					ret = ENCODE_SJIS;			/* default */
				}
			}
			break;
	}
	return(ret);
}


/*-------------------------------------------------------------------------------
Name    :	CheckOnSJIS
Desc    :	SHIFT-JISR[h̉\邩`FbN
Param   :	pos		
			btm		̖
Return  :	_
Note    :	
-------------------------------------------------------------------------------*/
static int CheckOnSJIS(BYTE *pos, BYTE *btm)
{
	int		fstOnTwo;
	int		point;

	fstOnTwo = NO;
	point = 100;
	while((point > 0) && (pos < btm) && (*pos != '\n'))
	{
		if(fstOnTwo == YES)
		{
			if((*pos < 0x40) || (*pos > 0xFC))	/* 2oCgڂ 0x40`0xFC */
			{
				point = 0;
			}
			fstOnTwo = NO;
		}
		else if(*pos >= 0x81)
		{
			if((*pos < 0xA0) || (*pos > 0xDF))	/* pJiłȂ */
			{
				if(*pos >= 0xEB)		/* 1oCgڂ0xEBȍ~͂قƂǖ */
				{
					point -= 50;
				}
				fstOnTwo = YES;
			}
		}
		pos++;
	}
	if(fstOnTwo == YES)		/* PoCgڂŏIĂ̂͂  */
	{
		point = 0;
	}
	return(point);
}


/*-------------------------------------------------------------------------------
Name    :	CheckOnEUC
Desc    :	EUCR[h̉\邩`FbN
Param   :	pos		
			btm		̖
Return  :	_
Note    :	
-------------------------------------------------------------------------------*/
static int CheckOnEUC(BYTE *pos, BYTE *btm)
{
	int		fstOnTwo;
	int		point;

	fstOnTwo = 0;
	point = 100;
	while((point > 0) && (pos < btm) && (*pos != '\n'))
	{
		if(fstOnTwo == 1)
		{
			if((*pos < 0xA1) || (*pos > 0xFE))	/* 2oCgڂ 0xA1`0xFE */
			{
				point = 0;
			}
			fstOnTwo = 0;
		}
		else if(fstOnTwo == 2)		/* pJi */
		{
			if((*pos < 0xA0) || (*pos > 0xDF))	/* 2oCgڂ 0xA0`0xDF */
			{
				point = 0;
			}
			fstOnTwo = 0;
		}
		else
		{
			if(*pos == 0x8E)		/* 0x8E??͔pJi */
			{
				fstOnTwo = 2;
			}
			else if((*pos >= 0xA1) && (*pos <= 0xFE))
			{
				fstOnTwo = 1;
			}
		}
		pos++;
	}
	if(fstOnTwo != 0)		/* PoCgڂŏIĂ̂͂  */
	{
		point = 0;
	}
	return(point);
}


/*-------------------------------------------------------------------------------
Name    :	CheckOnUTF8
Desc    :	UTF8̉\邩`FbN
Param   :	pos		
			btm		̖
Return  :	_
Note    :	
-------------------------------------------------------------------------------*/
static int CheckOnUTF8(BYTE *pos, BYTE *btm)
{
	int		rest;
	int		point;

	rest = 0;
	point = 100;
	while((point > 0) && (pos < btm) && (*pos != '\n'))
	{
		if(rest > 0)
		{
			point += 10;
			if((*pos & 0xC0) != 0x80)
			{
				point = 0;
			}
			rest--;
		}
		else if((*pos & 0x80) == 0x00)
		{
			rest = 0;
		}
		else if((*pos & 0xE0) == 0xC0)
		{
			rest = 1;
		}
		else if((*pos & 0xF0) == 0xE0)
		{
			rest = 2;
		}
		else if((*pos & 0xF8) == 0xF0)
		{
			rest = 3;
		}
		else if((*pos & 0xFC) == 0xF8)
		{
			rest = 4;
		}
		else if((*pos & 0xFE) == 0xFC)
		{
			rest = 5;
		}
		else
		{
			point = 0;
		}
		pos++;
	}
	return(point);
}


/*-------------------------------------------------------------------------------
Name    :	CheckTerminationCode
Desc    :	sR[h𒲂ׂ
Param   :	text		
			size		oCg
Return  :	sR[h (TERM_xxxx)
Note    :	
-------------------------------------------------------------------------------*/
int CheckTerminationCode(LPTSTR text, int size)
{
	LPTSTR	pos;
	int		ret;

	ret = TERM_UNKNOWN;
	pos = wmemchr(text, _T('\r'), size);
	if(pos != NULL)
	{
		if(((pos - text) < (size - 1)) && (*(pos+1) == _T('\n')))
		{
			ret = TERM_CRLF;
		}
		else
		{
			ret = TERM_CR;
		}
	}
	else if(wmemchr(text, _T('\n'), size) != NULL)
	{
		ret = TERM_LF;
	}
	return(ret);
}


